/* 
 * Copyright 2002 Vinicius da Silveira Serafim <serafim@inf.ufrgs.br>
 *                Alvaro Reguly <alvaro@reguly.net>
 *
 * This file is part of SOFFIC.
 *
 * SOFFIC is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SOFFIC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with SOFFIC; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * Changelog format
 *
 * [Month Year Your_Name] what_was_changed
 */

/*
 * [Feb 2002 V. Serafim ] First release.
 */


#include "sofficadm.h"


/*
 *	digest_getid
 *
 *	Just make the conversion of soffic constants to
 *	openssl digest id.
 */
static const EVP_MD *digest_getid(int digestid)
{
	switch (digestid) {
		case SOF_DIGEST_MD5:
			return (EVP_get_digestbyname("md5"));

		case SOF_DIGEST_SHA1:
			return (EVP_get_digestbyname("sha1"));
	}

	fprintf(stderr, "digest file: unknown digest algoritm\n");
	return NULL;
}


/*
 *	digest_file
 *
 *	Calculates the hash for a given file.
 */
int digest_file(int digestid, char *filename, char *digest)
{
	int i;
	FILE *fp;

	EVP_MD_CTX mdctx;
	int blocksize;
	const EVP_MD *md;
	u_char output[EVP_MAX_MD_SIZE];

	u_char buffer[SOF_BUFFSIZE];
	int bytesread = SOF_BUFFSIZE;

	OpenSSL_add_all_digests();

	md = digest_getid(digestid);

	if (!md) {
		printf("BUG! No such digest function\n");
		return -1;
	}


	EVP_DigestInit(&mdctx, md);

	if ((fp=fopen(filename, "r")) == NULL)
		return -1;

	while (bytesread == SOF_BUFFSIZE) {
		bytesread = fread(buffer, 1, SOF_BUFFSIZE, fp);
		if (bytesread > 0)
			EVP_DigestUpdate(&mdctx, buffer, bytesread);
	}
	
	fclose(fp);

	EVP_DigestFinal(&mdctx, output, &blocksize);

	memset(digest, 0, SOF_DIGEST_MAXSZ);
	for (i=0; i<blocksize; i++)
		digest[i] = output[i];

	return 0;
}

// EOF
