/* 
 * Copyright 2002 Vinicius da Silveira Serafim <serafim@inf.ufrgs.br>
 *                Alvaro Reguly <alvaro@reguly.net>
 *
 * This file is part of SOFFIC.
 *
 * SOFFIC is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SOFFIC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with SOFFIC; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * Changelog format
 *
 * [Month Year Your_Name] what_was_changed
 */

/*
 * [Feb 2002 V. Serafim ] First release.
 */


#include "sofficadm.h"

typedef struct {
        int  open;
	int  options;
	char name[SOF_BUFFSIZE];
	int  recursive; /* -1 means not set, 0 set to OFF, >0 set to on */
	int  digest;
	int  desc;
} _group;
	
typedef struct {
	int groupid;
	int open;
	int not;
        char name[SOF_BUFFSIZE];
        int options;
	int recursive;	/* -1 means not set, 0 set to OFF, >0 set to on */
	int digest;
} _object;

typedef struct {
	int dirs;
	int processed;
	int duplicated;
	int errors;
	int valid;
} _statistics;

enum { GROUP, OBJECT };

/* global variables declarations */

_statistics total = { 0, 0, 0, 0, 0 };

int digest_algorithm = 0;

/* this is the first list created from the description file */
int objectcount = 0;
_object **objectlist = NULL;

/* this is the final list created from the objectlist*/
int hlcount = 0;
_hl_entry **hashlist = NULL;

int groupid = 0;	/* this is just a counter */
_group group = { 0, 0, "", 0, SOF_DIGEST_MD5, 0 };
_object object = { 0, 0, 0, "", 0, -1, SOF_DIGEST_NONE };


/*
 *	hl_exit_help
 *
 *	Print usage information for -H.
 */
void hl_exit_help(char *argv0) {

	printf("usage: %s -H [options]\n\n",argv0);

	printf("  -g\tGenerate\n");
	printf("\t-d <md5/sha1>\t\tdigest algorithm (default md5)\n");
	printf("\t-i <file>\t\tinput file (must be specified)\n");
	printf("\t-o <file>\t\toutput file (default " SOF_HL_FILE ")\n");
	printf("\n");
	printf("  -l\tList contents\n");
	printf("\t-i <file>\t\tinput file (default " SOF_HL_FILE ")\n");
	printf("\t-o <file>\t\toutput file (default stdout)\n");
	printf("\n");

	exit(0);
}


/*
 *	hl_file2mem
 *
 *	Create hashlist[] using HL entries from a file.
 */
static void hl_file2mem(char *hlfile) {

	FILE *fp;
	int newpos;
	_hl_entry newentry;

	if ((fp = fopen(hlfile, "r")) == NULL) {
		fprintf(stderr, __FUNCTION__": cannot open %s\n", hlfile);
		perror(__FUNCTION__);
		exit(2);
	}

	while (1) {
		/* read exactly one entry at a time */
		if (fread(&newentry, sizeof(_hl_entry), 1, fp) != 1) {
			if (feof(fp))
				break;
			else {
				/* ops, we got an invalid sized entry */
				fprintf(stderr,"HL File inconsistency: %s\n", hlfile);
				exit(2);
			}
		}

		/* insert at the hashlist */
		hashlist = (_hl_entry **) misc_array_realloc((void **)hashlist, \
			sizeof(_hl_entry *), sizeof(_hl_entry), &hlcount, 1);
		newpos = hlcount - 1;
		memcpy(hashlist[newpos], &newentry, sizeof(_hl_entry));
	}

	fclose(fp);
}


/*
 *	hl_mem2file
 *
 *	Writes hashlist[] to the output file.
 */
static void hl_mem2file(char *file) {

	int i;
	FILE *fp = NULL;

	/* generate the output file */
	if ((fp = fopen(file, "w")) == NULL) {
		fprintf(stderr, __FUNCTION__": cannot open %s\n", file);
		perror(__FUNCTION__);
		exit(2);
	}

	i = 0;
	while (i < hlcount) {
		if (fwrite(hashlist[i], sizeof(_hl_entry), 1, fp) != 1) {
			fprintf(stderr,__FUNCTION__": error on writing "
							" to %s\n", file);
			exit(2);
		}
		i++;
	}

	fclose(fp);
}


/*
 *	hl_find_newpos
 *
 *	Create an hashlist position for a newentry.
 */
static int hl_find_newpos(_hl_entry *newentry) {

	int i = 0, j = 0;

	if (hlcount == 0) {
		hashlist = (_hl_entry **) misc_array_realloc((void **)hashlist, \
			sizeof(_hl_entry *), sizeof(_hl_entry), &hlcount, 1);
		return 0;
	}

	/* this is just a sequential lookup */
	while (i < hlcount) {
		/* check for an already included inode */
		if ((hashlist[i]->dev == newentry->dev) &&
		    (hashlist[i]->ino == newentry->ino)) {
			total.duplicated++;
			return -1;
		}

		if ((hashlist[i]->dev == newentry->dev) &&
		    (hashlist[i]->ino > newentry->ino))
			break;

		if (hashlist[i]->dev > newentry->dev)
			break;

		i++;
	}

	/* create an empty node */
	hashlist = (_hl_entry **) misc_array_realloc((void **)hashlist, \
		sizeof(_hl_entry *), sizeof(_hl_entry), &hlcount, 1);
	
	j = hlcount - 1;	/* the index of the empty node */
	while (j != i) {
		memcpy(hashlist[j], hashlist[j-1], sizeof(_hl_entry));
		j--;
	}

	return i;
}


/*
 *	hl_insert_node
 *
 *	Insert a new node to hashlist.
 */
static int hl_insert_node(_object *obj, struct stat *filestat)
{
	long int flags = 0;
	int newpos, error;
	_hl_entry newentry;

	if (obj->digest == SOF_DIGEST_MD5)
		BF_CLR(flags, SOF_HLBF_SHA1);
	else if (obj->digest == SOF_DIGEST_SHA1)
		BF_SET(flags, SOF_HLBF_SHA1);
	else {
		printf("bug!, unknown digest algorithm\n");
		exit(2);
	}
		
	newentry.flags	= flags;
	newentry.dev	= filestat->st_dev;
	newentry.ino	= filestat->st_ino;
	memset(newentry.hash, 0, SOF_DIGEST_MAXSZ);

	error = digest_file(obj->digest, obj->name, newentry.hash);
	if (error) {
		fprintf(stderr,__FUNCTION__": no digest for %s\n", obj->name);
		total.errors++;
		return -1;
	}

	newpos = hl_find_newpos(&newentry);
	if (newpos != -1) {
		memcpy(hashlist[newpos], &newentry, sizeof(_hl_entry));
		total.valid++;
	} else
		return -1;

	return 0;
}


/*
 * 	hl_search_object
 *
 * 	Searches for a more accurate object description.
 */
int hl_search_object(_object *obj)
{
	int i = 0, ret = 0;

	/* find the first object in the same group of obj */
	while ((i<objectcount) && (objectlist[i]->groupid != obj->groupid))
		i++;

	while ((i<objectcount) && (objectlist[i]->groupid == obj->groupid)) {
		ret = strcmp(obj->name, objectlist[i]->name);
		if (ret == 0)	/* a match */
			return 1;
		/* if they difer in one char, maybe the difference is just a slash */
		else if (((strlen(obj->name) - strlen(objectlist[i]->name)) == 1) && \
			 (obj->name[strlen(obj->name)-1] == '/'))
			return 1;
		i++;
	}

	return 0;
}


/*
 *	hl_process_object
 *
 *	Take an object and fill the hashlist.
 */
void hl_process_object(_object *obj)
{
	struct stat *filestat = NULL;
	_object objdir;
        DIR *dir;
        struct dirent *dirent;

	total.processed++;

	if (obj->not)
		goto out;
	
        filestat = misc_getfilestat(obj->name);
        if (!filestat) {
                total.errors++;
                goto out;
        }

	/* by default links are not followed */
	if (S_ISLNK(filestat->st_mode))
		goto out;

	/* a regular file, just insert it in the hashlist */
	if (S_ISREG(filestat->st_mode))
		hl_insert_node(obj, filestat);

	else if (S_ISDIR(filestat->st_mode)) {

		if  (!obj->recursive)
			goto out;

		printf("\t%s\n", obj->name);
		total.dirs++;

		if ((dir = opendir(obj->name)) == NULL) {
	                perror(__FUNCTION__);
			total.errors++;
        	        goto out;
		}

		obj->recursive--;
		while ((dirent = readdir(dir)) != NULL) {
			if (dirent->d_name[0] == '.')
				continue;

			/* copy the original object */
			memcpy(&objdir, obj, sizeof(_object));

			/* and complete the new object name */
			if (objdir.name[strlen(objdir.name)-1] != '/')
				strncat(objdir.name, "/", 1);
			strncat(objdir.name, dirent->d_name, SOF_BUFFSIZE - strlen(objdir.name));

			if (!hl_search_object(&objdir))
				hl_process_object(&objdir);
		}
		closedir(dir);
		obj->recursive++;
	}

out:
	free(filestat);
	return;
}


/*
 *	hl_parse_getoption
 *	
 *	Read an option name and value from a line.
 */
int hl_parse_getoption(char *line, int linelen, int *i, int who)
{
	int j, old_i = *i;
	char option[SOF_BUFFSIZE] = "";
	char value[SOF_BUFFSIZE]  = "";
	int  setto = 0;
	char endchars[] = "\t\n),= ";

	/* get the option name */
	j=0;
	while ((*i<linelen) && (strchr(endchars, line[*i]) == NULL))
		option[j++] = line[(*i)++];

	option[j] = '\0';
	
	/* get the option value, if given */
	j=0;
	if ((option[0] != '\0') && (line[*i] == '=')) {
		(*i)++;
		while ((*i<linelen) && (strchr(endchars, line[*i]) == NULL))
			value[j++] = line[(*i)++];

		value[j] = '\0';
	}

	/* Test & Set */

	/* recursive */
	if (strcmp("recursive", option) == 0) {
		if (strcmp("on",value) == 0)
			setto = 1;
		else if (strcmp("off",value) == 0)
			setto = 0;
		else if (isdigit(value[0]))
			setto = atoi(value);
		else
			goto error_value;	/* bad option */

		if (who == GROUP)
			group.recursive = setto;
		else
			object.recursive = setto;

		return 0;		/* ok, everything is tested and set */
	}

	/* digest */
	if (strcmp("digest", option) == 0) {
		if (strcmp("md5", value) == 0)
			setto = SOF_DIGEST_MD5;
		else if (strcmp("sha1", value) == 0)
			setto = SOF_DIGEST_SHA1;
		else
			goto error_value;

		if (who == GROUP)
			group.digest = setto;
		else
			object.digest = setto;

		return 0;
	}

	printf("error: %s=%s\n", option, value);
	(*i) = old_i;
	return -1;
	
error_value:
	printf("error: %s=%s\n", option, value);
	(*i)-=strlen(value);
	return -1;
}

/*
 *	hl_parse
 *
 *	Process one input line.
 */
int hl_parse(char *line, int linelen) {

	int i=0, j=0;

	/* close a possibly open object */
	object.groupid	 = 0;
	object.open      = 0;
	object.not       = 0;
	object.name[0]   = '\0';
	object.options   = 0;
	object.recursive = -1;
	object.digest    = SOF_DIGEST_NONE;

	while (i<linelen) {

		switch (line[i]) {


		case ' ':
		case ',':	/* options separator */
		case '\t':	/* tabs */
			break;
		
		case '=':	/* what this is doing here? */
			return -i;
			
		case '\n': 	/* end of the line */
		case '#':	/* commend */
			/* ok, we have an object to be processed */
			if (object.open && !object.options) {
				object.groupid = groupid;
				return 1;
			}

			/* do nothing */
			if (!group.options && !object.options) return 0;

			/* syntax error */
			return -i;
		
		case '(':
			/* a new group? */
			if (!group.open) {
				group.open = 1;
				group.options = 1;
				groupid++;
				break;
			}

			/* object options? */
			if (object.open && (object.name[0] != '\0') &&
				!object.options) {
				object.options = 1;
				break;
			}

			/* none of them! syntax error */
			return -i;
			
		case ')':
			/* closing an object option? */
			if (object.open && object.options) {
				object.options = 0;
				break;
			}

			/* closing a group option? */
			if (group.open && group.options) {
				group.options = 0;
				break;
			}

			/* syntax error */
			return -i;

		case '{': /* open a group description */
			if (group.open && !group.options && !group.desc) {
				group.desc = 1;
				break;
			}

			/* syntax error */
			return -i;

		case '}': /* closing a group description */
			if (group.open && group.desc && !object.options) {
				group.desc      = 0;
				group.open      = 0;
				group.options   = 0;
				group.name[0]   = '\0';
				group.recursive = 0;
				group.digest    = SOF_DIGEST_MD5;
				break;
			}

			/* syntax error */
			return -i;

		case '!': /* negative signal for an object */
			if (group.desc && !object.open) {
				object.not = 1;
				break;
			}

			/* syntax error */
			return -i;

		default:
			/* a new object? */
			if (group.desc && !object.open && !object.options) {
				j = 0;
				while ((i<linelen) && (strchr("\t\n( ",line[i]) == NULL))
					object.name[j++] = line[i++];
				
				object.name[j] = '\0';
				object.open = 1;
				i--;
				break;
			}

			/* object options */
			if (object.options) {
				if (hl_parse_getoption(line, linelen, &i, OBJECT) != 0)
					return -(i--);
				else {
					i--;
					break;
				}
			}

			/* the name for a group? */
			if ((group.options) && (group.name[0] == '\0')) {
				j=0;
				while ((i<linelen) && (strchr("\t\n), ",line[i]) == NULL))
					group.name[j++] = line[i++];

				group.name[j] = '\0';
				i--;
				break;
				
			}
			
			/* group options */
			if ((group.options) && (group.name[0] != '\0')) {
				if (hl_parse_getoption(line, linelen, &i, GROUP) != 0)
					return -(i--);
				else {
					i--;
					break;
				}
			}

			return -i;
			
		} /* switch */

		i++;

	} /* while */

	return 0;
}


/*
 *	hl_generate
 *
 *	Generate a new hash list.
 */
static void hl_generate(char **files) {

	int i, ret, lineno = 0;
	FILE *fp = NULL;
	char buffer[SOF_BUFFSIZE];

	printf("Generating %s...\n", files[HL_FILE_OUT]);

	if ((fp = fopen(files[HL_FILE_IN], "r")) == NULL) {
		perror(__FUNCTION__);
		exit_error("cannot open input file\n");
	}

	/* 1st: read parse the description file and populate objectlist */
	while (fgets(buffer, SOF_BUFFSIZE, fp) != NULL) {

		lineno++;

		/* process a line */
		ret = hl_parse(buffer, strlen(buffer));
		
		if (ret == 0)	/* just go to the next line */
			continue;

		if (ret == 1) { /* ops, we've got an object */

			/* set unset options */
			if (object.recursive == -1)
				object.recursive = group.recursive;
			if (object.digest == SOF_DIGEST_NONE)
				object.digest = group.digest;
		
			/* add one element to the object list */
			objectlist = (_object **) misc_array_realloc((void **)objectlist, \
					sizeof(_object *), sizeof(_object), &objectcount, 1);

			memcpy(objectlist[objectcount-1], &object, sizeof(_object));
			
			continue; /* next line */
		}
			
		if (ret < 0) {	/* print syntax error message */
			printf("syntax error on line %d\n%s", lineno, buffer);
			if (buffer[strlen(buffer)-1] != '\n')
				printf("\n");
			for(i=0; i<-ret; i++) {
				if (buffer[i] == '\t')
					printf("\t");
				else
					printf(" ");
			}
			printf("^\n");
			exit(1);
		}

	} /* while */

	fclose(fp);


	/* 2nd: build the HashList */
	i=0;
	while (i<objectcount) {
		hl_process_object(objectlist[i]);
		i++;
	}

	goto end;	
		
		/* Take off \n */

		buffer[strlen(buffer)-1] = '\0';

		switch (ret) {

		case 0:			/* ignore */
			break;

		case 1:			/* normal group */
/*			printf("-> %s\n", buffer);
			flags = 0;
			break;
*/
		case 2:			/* special group */
/*			if (strncmp(buffer, "COREFILE", 8) == 0)
				BF_SETTO(flags, SOF_HLBF_CORE);
			else if (strncmp(buffer, "CONFIGFILE", 10) == 0)
				BF_SETTO(flags, SOF_HLBF_CONF);
			else if (strncmp(buffer, "TFL", 3) == 0)
				BF_SETTO(flags, SOF_HLBF_TFL);
			else if (strncmp(buffer, "SOFFICADM", 3) == 0)
				BF_SETTO(flags, SOF_HLBF_ADM);
			else
				exit_error("sintax error in input file\n");

			printf("-> %s\n", buffer);
			break;
*/
		case 3:			/* normal entry */
/*			hl_process_entry(buffer, 0, flags);
			break;
*/
		case 4:			/* recursive entry */
/*			hl_process_entry(buffer, 1, flags);
			break;
*/
		case -1:
		default:
/*			exit_error("sintax error in input file\n");
*/
		}

end:

	hl_mem2file(files[HL_FILE_OUT]);

//	fclose(fp);

	printf("\nOperation resume:\n");
	printf("processed items         %5d\n", total.processed);
	printf("directories             %5d\n", total.dirs);
	printf("duplicated              %5d\n", total.duplicated);
	printf("errors                  %5d\n", total.errors);
	printf("valid items             %5d\n", total.valid);
	printf("generated file: %s\n",  files[HL_FILE_OUT]);
}

/*
 *	hl_list
 *
 *	List the HL entries
 */
static void hl_list(char *hlfile) {

	int i, j;

	hl_file2mem(hlfile);
	
	for(i=0; i<hlcount; i++) {
		printf("%d;%ld;", (int) hashlist[i]->dev, 
			hashlist[i]->ino);

		for(j=0; j<SOF_DIGEST_MAXSZ; j++)
			printf("%02x", hashlist[i]->hash[j]);

		printf(";%08lx\n", hashlist[i]->flags);
	}
}


/*
 *	hl
 *
 *	The Main HL function
 */
void hl(int opr, int dgst, char **files) {

	digest_algorithm = dgst;

	switch (opr) {
		case HL_GENERATE:
			if (files[HL_FILE_OUT][0] == '\0')
				strncpy(files[HL_FILE_OUT], SOF_HL_FILE, SOF_BUFFSIZE);
			if (files[HL_FILE_IN][0] == '\0')
				exit_usehelp();

			hl_generate(files);
			break;

		case HL_LIST:
			if (files[HL_FILE_IN][0] == '\0')
				strncpy(files[HL_FILE_IN], SOF_HL_FILE, SOF_BUFFSIZE);
			hl_list(files[HL_FILE_IN]);
			break;

		default:
			exit_usehelp();
	}
}

