/* 
 * Copyright 2002 Vinicius da Silveira Serafim <serafim@inf.ufrgs.br>
 *                Alvaro Reguly <alvaro@reguly.net>
 *
 * This file is part of SOFFIC.
 *
 * SOFFIC is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SOFFIC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with SOFFIC; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * Changelog format
 *
 * [Month Year Your_Name] what_was_changed
 */

/*
 * [Feb 2002 V. Serafim ] First release.
 */


#include "sofficadm.h"

void exit_error(char *message)
{

	fprintf(stderr,"ERROR: %s", message);
	exit(1); 
}

/*
 *	exit_usehelp
 *
 *	Just tell the user to try --help and exit.
 */
void exit_usehelp(void)
{
        fprintf(stderr, "Parameter error! Try --help\n");
        exit(2);
}


/*
 * 	misc_getfilestat
 *
 * 	Get the stat for a file.
 */
struct stat *misc_getfilestat(char *filename)
{
	struct stat *filestat;

	filestat = (struct stat *) malloc(sizeof(struct stat));

	if (!filestat) {
		fprintf(stderr, __FUNCTION__": memory allocation error\n");
		return NULL;
	}

        /* from the lstat man page:
	   "lstat  is  identical to stat, except in the case of a symbolic link,
	   where the link itself is stat-ed, not the file that it refers to." */
	if (lstat(filename, filestat) != 0) {
		perror(filename);
		free(filestat);
		return NULL;
	}

	return filestat;
}


/*
 * 	misc_array_realloc
 *
 * 	Allocate memory for an array.
 */
void **misc_array_realloc(void **array, int pointer_size, int size, int *elements, int addnum)
{
	int i;
	long new_size = ((*elements)+addnum) * pointer_size;

	array = (void **) realloc(array, new_size);
		
	if (!array) {
		fprintf(stderr,__FUNCTION__": memory allocation error\n");
		exit(2);
	}

	for (i=*elements; i<((*elements)+addnum); i++) {
		array[i] = (void *) malloc(size);
		if (!array[i]) {
			fprintf(stderr,__FUNCTION__": memory allocation error\n");
			exit(2);
		}
		memset(array[i], 0, size);
	}
	
	(*elements)+=addnum;

	return array;
}











// EOF
