/* 
 * Copyright 2002 Vinicius da Silveira Serafim <serafim@inf.ufrgs.br>
 *                Alvaro Reguly <alvaro@reguly.net>
 *
 * This file is part of SOFFIC.
 *
 * SOFFIC is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SOFFIC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with SOFFIC; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * Changelog format
 *
 * [Month Year Your_Name] what_was_changed
 */

/*
 * [Feb 2002 V. Serafim ] First release.
 */


#include "sofficadm.h"

enum { CMD_NONE, CMD_FLAGS, CMD_HASHLIST };

/*
 *	exit_usage
 *
 *	Show usage information and exit.
 */
void exit_usage(char *argv0)
{
	printf("usage: %s [command] [options]\n\n", argv0);

	printf("\t-H\tHash List operations.\n");
	printf("\n\t--version\tShow only version information.\n");
	printf("\t--help\t\tShow this help message.\n");

	printf("\nFor more detailed help use: %s [command] --help\n\n", argv0);

	exit(0);
}


int main(int argc, char **argv)
{

	int i, opt, cmd = CMD_NONE, opr = CMD_NONE, option_index = 0;

	static char optstring[] = "Hgld:a:i:o:";
	static struct option long_options[] = {
		{"version", 0, 0 , 1},
		{"help", 0, 0, 2},
		{0, 0, 0, 0}
	};

	int digest = SOF_DIGEST_MD5;
	char **files;

	files = (char **) malloc(sizeof(char *) * 3);
	for (i=0; i<3; i++) {
		files[i] = (char *) malloc(sizeof(char) * SOF_BUFFSIZE);
		files[i][0] = '\0';
	}

	/* our presentation string :) */
	printf("SOFFIC administration tool v%s, ", SOF_SOFFICADM_VERSION);
	printf("compiled for SOFFIC v%s\n", SOF_VERSION);
	
	/* set umask -rw------ */
	umask(0177);

	/* check the user id */
	if (getuid() != 0) {
		fprintf(stderr, "You must be root to use sofficadm!\n");
		exit(1);
	} 

	 /* Parse command line arguments
	    Lets get what the user want from us */
	while (1) {
		opt = getopt_long(argc, argv, optstring,
			long_options, &option_index);
		if (opt == -1) break;

		switch (opt) {

		case 1: /* version */
			exit(0);

		case 2: /* help */
			if (cmd == CMD_HASHLIST) hl_exit_help(argv[0]);
			exit_usage(argv[0]);

		case 'H': /* HashList */
			if (cmd != CMD_NONE) exit_usehelp();
			cmd = CMD_HASHLIST;
			break;

		case 'l': /* list */
			if ((cmd == CMD_HASHLIST) && (opr == HL_NONE))
				{ opr = HL_LIST; break; }

			exit_usehelp();

		case 'g': /* generation */
			if (cmd != CMD_HASHLIST) exit_usehelp();
			if ((cmd == CMD_HASHLIST) && (opr != HL_NONE))
				exit_usehelp();
			opr = HL_GENERATE;
			break;

		case 'd': /* digest name */
			if (cmd != CMD_HASHLIST) exit_usehelp();
			if (strncmp(optarg, "md5", 3) == 0)
				digest = SOF_DIGEST_MD5;
			else if (strncmp(optarg, "sha1", 4) == 0)
				digest = SOF_DIGEST_SHA1;
			else 
				hl_exit_help(argv[0]);

		case 'i': /* input file */
			if (cmd != CMD_HASHLIST) exit_usehelp();

			strncpy(files[HL_FILE_IN], optarg, SOF_BUFFSIZE);
			break;

		case 'o': /* output file */
			if ((cmd != CMD_HASHLIST) || (opr != HL_GENERATE))
				exit_usehelp();

			strncpy(files[HL_FILE_OUT], optarg, SOF_BUFFSIZE);
			break;

		default:
			exit_usehelp();

		}
	}

	if (cmd == CMD_NONE) exit_usehelp();

	if (cmd == CMD_HASHLIST)
		if (opr == CMD_NONE) exit_usehelp();

	/* lets call who dows the real work */
	switch (cmd) {
		case CMD_HASHLIST:
			hl(opr, digest, files);
			break;
	}

	exit(0);
}

// EOF
